__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os
import numpy as np
import shutil


from .. import TSNEEncoder
from jina.executors import BaseExecutor

requires_train_after_load = False
input_dim = 28
target_output_dim = 2

def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)

def test_encoder():
    encoder = TSNEEncoder(output_dim=target_output_dim)
    assert encoder is not None
    test_data = np.random.rand(10, input_dim)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (test_data.shape[0], target_output_dim)
    assert type(encoded_data) is np.ndarray

def test_save_and_load():
    encoder = TSNEEncoder(output_dim=target_output_dim)
    assert encoder is not None
    test_data = np.random.rand(10, input_dim)
    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)

    if not requires_train_after_load:
        # some models are not deterministic when training, so even with same training data, we cannot ensure
        # same encoding results
        encoded_data_test = encoder_loaded.encode(test_data)
        np.testing.assert_array_equal(
            encoded_data_test, encoded_data_control)
    rm_files([encoder.save_abspath])


def test_save_and_load_config():
    encoder = TSNEEncoder(output_dim=target_output_dim)
    assert encoder is not None
    encoder.save_config()
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    test_data = np.random.rand(10, input_dim)
    encoded_data_test = encoder_loaded.encode(test_data)
    assert encoded_data_test.shape == (10, target_output_dim)
    rm_files([encoder.save_abspath, encoder.config_abspath])


__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from . import default_logger
from .queue import __sse_queue__, __profile_queue__
from .. import JINA_GLOBAL, __version__
from ..helper import yaml


def start_sse_logger(server_config_path: str, flow_yaml: str = None):
    """Start a logger that emits server-side event from the log queue, so that one can use a browser to monitor the logs

    :param host: host address of the server
    :param port: port of the server
    :param endpoint_log: endpoint for the log
    :param endpoint_yaml: endpoint for the yaml

    Example:

    .. highlight:: javascript
    .. code-block:: javascript

        var stream = new EventSource('http://localhost:5000/log/stream');
        stream.onmessage = function (e) {
            console.info(e.data);
        };
        stream.onerror = function (err) {
            console.error("EventSource failed:", err);
            stream.close()
        };

    """
    try:
        from flask import Flask, Response, jsonify
        from flask_cors import CORS
        from gevent.pywsgi import WSGIServer
        import gevent
    except ImportError:
        raise ImportError('Flask or its dependencies are not fully installed, '
                          'they are required for serving HTTP requests.'
                          'Please use pip install "jina[http]" to install it.')

    try:
        with open(server_config_path) as fp:
            _config = yaml.load(fp)
    except Exception as ex:
        default_logger.error(ex)
    JINA_GLOBAL.logserver.address = f'http://{_config["host"]}:{_config["port"]}'

    JINA_GLOBAL.logserver.ready = JINA_GLOBAL.logserver.address + \
                                  _config['endpoints']['ready']
    JINA_GLOBAL.logserver.shutdown = JINA_GLOBAL.logserver.address + \
                                     _config['endpoints']['shutdown']

    app = Flask(__name__)
    CORS(app)
    server = WSGIServer((_config['host'], _config['port']), app, log=None)

    def _log_stream():
        while True:
            try:
                gevent.sleep(0)
                message = __sse_queue__.get()
                yield f'data: {message.msg}\n\n'
            except EOFError:
                yield 'LOG ENDS\n\n'
                break

    def _profile_stream():
        while True:
            try:
                gevent.sleep(0)
                message = __profile_queue__.get()
                yield f'data: {message.msg}\n\n'
            except EOFError:
                yield 'PROFILE ENDS\n\n'
                break

    @app.route(_config['endpoints']['log'])
    def get_log():
        """Get the logs, endpoint `/log/stream`  """
        return Response(_log_stream(), mimetype="text/event-stream")

    @app.route(_config['endpoints']['yaml'])
    def get_yaml():
        """Get the yaml of the flow  """
        return flow_yaml

    @app.route(_config['endpoints']['profile'])
    def get_profile():
        """Get the profile logs, endpoint `/profile/stream`  """
        return Response(_profile_stream(), mimetype='text/event-stream')

    @app.route(_config['endpoints']['podapi'])
    def get_podargs():
        """Get the default args of a pod"""

        from jina.main.parser import set_pod_parser
        from argparse import _StoreAction, _StoreTrueAction
        port_attr = ('help', 'choices', 'default')
        d = {}
        parser = set_pod_parser()
        for a in parser._actions:
            if isinstance(a, _StoreAction) or isinstance(a, _StoreTrueAction):
                d[a.dest] = {p: getattr(a, p) for p in port_attr}
                if a.type:
                    d[a.dest]['type'] = a.type.__name__
                elif isinstance(a, _StoreTrueAction):
                    d[a.dest]['type'] = 'bool'
                else:
                    d[a.dest]['type'] = a.type

        d = {'pod': d, 'version': __version__, 'usage': parser.format_help()}
        return jsonify(d)

    @app.route(_config['endpoints']['shutdown'])
    def shutdown():
        server.stop()
        return 'Server shutting down...'

    @app.route(_config['endpoints']['ready'])
    def is_ready():
        return Response(status=200)

    try:
        server.serve_forever()
    except Exception as ex:
        default_logger.error(ex)



import os
import pytest
import numpy as np

from .. import Wav2VecSpeechEncoder


@pytest.mark.skipif('JINA_TEST_PRETRAINED' not in os.environ, reason='skip the pretrained test if not set')
def test_encoding_results():
    target_output_dim = 512
    batch_size = 10
    signal_length = 1024
    test_data = np.random.randn(batch_size, signal_length).astype('f')
    encoder = Wav2VecSpeechEncoder(model_path='/tmp/wav2vec_large.pt', input_sample_rate=16000)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape[0] == batch_size
    assert encoded_data.shape[1] % target_output_dim == 0



from jina.flow import Flow

from .. import Sentencizer


def test_sentencier_en():
    sentencizer = Sentencizer()
    text = 'It is a sunny day!!!! When Andy comes back, we are going to the zoo.'
    crafted_chunk_list = sentencizer.craft(text, 0)
    assert len(crafted_chunk_list) == 2


def test_sentencier_en_new_lines():
    """
    New lines are also considered as a separator.
    """
    sentencizer = Sentencizer()
    text = 'It is a sunny day!!!! When Andy comes back,\n' \
           'we are going to the zoo.'
    crafted_chunk_list = sentencizer.craft(text, 0)
    assert len(crafted_chunk_list) == 3


def test_sentencier_en_float_numbers():
    """
    Separators in float numbers, URLs, emails, abbreviations (like 'Mr.')
    are not taking into account.
    """
    sentencizer = Sentencizer()
    text = 'With a 0.99 probability this sentence will be ' \
           'tokenized in 2 sentences.'
    crafted_chunk_list = sentencizer.craft(text, 0)
    assert len(crafted_chunk_list) == 2


def test_sentencier_en_trim_spaces():
    """
    Trimming all spaces at the beginning an end of the chunks.
    Keeping extra spaces inside chunks.
    Ignoring chunks with only spaces.
    """
    sentencizer = Sentencizer()
    text = '  This ,  text is...  . Amazing !!'
    chunks = [i['text'] for i in sentencizer.craft(text, 0)]
    locs = [i['location'] for i in sentencizer.craft(text, 0)]
    assert chunks, ["This ,  text is..." == "Amazing"]
    assert text[locs[0][0]:locs[0][1]], '  This  ==   text is...'
    assert text[locs[1][0]:locs[1][1]] == ' Amazing'

    def validate(req):
        assert req.docs[0].chunks[0].text, 'This  ==   text is...'
        assert req.docs[0].chunks[1].text == 'Amazing'

    f = Flow().add(uses='!Sentencizer')
    with f:
        f.index_lines(['  This ,  text is...  . Amazing !!'], output_fn=validate, callback_on_body=True)


def test_sentencier_cn():
    sentencizer = Sentencizer()
    text = '今天是个大晴天！安迪回来以后，我们准备去动物园。'
    crafted_chunk_list = sentencizer.craft(text, 0)
    # Sentencizer does not work for chinese because string.printable does not contain Chinese characters
    assert len(crafted_chunk_list) == 0



import os
import numpy as np
import shutil
import mock
import pytest

from .. import FlairTextEncoder
from jina.executors import BaseExecutor
from jina.executors.metas import get_default_metas

target_output_dim = 100


def get_metas():
    metas = get_default_metas()
    if 'JINA_TEST_GPU' in os.environ:
        metas['on_gpu'] = True
    return metas


def rm_files(tmp_files):
    for file in tmp_files:
        if file and os.path.exists(file):
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.isdir(file):
                shutil.rmtree(file, ignore_errors=False, onerror=None)


test_data = np.array(['it is a good day!', 'the dog sits on the floor.'])


class MockEmbedding:
    pass


class MockDocumentEmbedding:
    def embed(self, sentences):
        return np.random.random((len(sentences), target_output_dim))


class MockSentence:
    @property
    def embedding(self):
        return np.random.random((1, target_output_dim))


def _test_encoding_results(*args, **kwargs):
    encoder = FlairTextEncoder(embeddings=('word:glove',), pooling_strategy='mean')
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (2, target_output_dim)
    rm_files([encoder.config_abspath, encoder.save_abspath])


@mock.patch('flair.device', return_value='')
@mock.patch('flair.embeddings.WordEmbeddings', return_value=MockEmbedding())
@mock.patch('flair.embeddings.DocumentPoolEmbeddings', return_value=MockDocumentEmbedding())
@mock.patch('flair.data.Sentence', return_value=MockSentence())
def test_encoding_results_local(*args, **kwargs):
    _test_encoding_results(*args, **kwargs)


@mock.patch('flair.device', return_value='')
@mock.patch('flair.embeddings.WordEmbeddings', return_value=MockEmbedding())
@mock.patch('flair.embeddings.DocumentPoolEmbeddings', return_value=MockDocumentEmbedding())
@mock.patch('flair.data.Sentence', return_value=MockSentence())
def test_save_and_load(*args, **kwargs):
    encoder = FlairTextEncoder(embeddings=('word:glove',), pooling_strategy='mean')
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)
    assert encoder_loaded.embeddings == encoder.embeddings
    rm_files([encoder.config_abspath, encoder.save_abspath])


@mock.patch('flair.device', return_value='')
@mock.patch('flair.embeddings.WordEmbeddings', return_value=MockEmbedding())
@mock.patch('flair.embeddings.FlairEmbeddings', return_value=MockEmbedding())
@mock.patch('flair.embeddings.DocumentPoolEmbeddings', return_value=MockDocumentEmbedding())
@mock.patch('flair.data.Sentence', return_value=MockSentence())
def test_save_and_load_config(*args, **kwargs):
    encoder = FlairTextEncoder(embeddings=('flair:news-forward',), pooling_strategy='mean')
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    assert encoder_loaded.max_length == encoder.max_length
    rm_files([encoder.config_abspath, encoder.save_abspath])


@pytest.mark.skipif('JINA_TEST_PRETRAINED' not in os.environ, reason='skip the pretrained test if not set')
def test_encoding_results(*args, **kwargs):
    _test_encoding_results(*args, **kwargs)



from .. import SlidingWindowSegmenter


def test_sliding_window_segmenter():
    window_size = 20
    step_size = 10
    sliding_window_segmenter = SlidingWindowSegmenter(
        window_size=window_size, step_size=step_size)
    text = 'It is a sunny day!!!! When Andy comes back, we are going to the zoo.'
    crafted_chunk_list = sliding_window_segmenter.craft(text, 0)
    assert len(crafted_chunk_list) == len(text) // step_size



import tempfile
import os
import shutil

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from jina.executors.metas import get_default_metas
from .. import CustomImageTorchEncoder
from jina.executors import BaseExecutor


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)


class ExampleNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 10, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(10, 16, 5)
        self.fc1 = nn.Linear(16 * 53 * 53, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 53 * 53)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


def get_encoder():
    metas = get_default_metas()
    if 'JINA_TEST_GPU' in os.environ:
        metas['on_gpu'] = True
    path = tempfile.NamedTemporaryFile().name
    model = ExampleNet()
    torch.save(model, path)
    return CustomImageTorchEncoder(model_path=path, layer_name='conv1', metas=metas)


def test_encoding_results():
    output_dim = 10
    input_dim = 224
    encoder = get_encoder()
    test_data = np.random.rand(2, 3, input_dim, input_dim)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (2, output_dim)
    rm_files([encoder.save_abspath, encoder.config_abspath, encoder.model_path])


def test_save_and_load():
    input_dim = 224
    encoder = get_encoder()
    test_data = np.random.rand(2, 3, input_dim, input_dim)
    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)
    encoded_data_test = encoder_loaded.encode(test_data)
    assert encoder_loaded.channel_axis == encoder.channel_axis
    np.testing.assert_array_equal(encoded_data_control, encoded_data_test)
    rm_files([encoder.save_abspath, encoder.config_abspath, encoder.model_path])



import gzip
import os
import shutil

import numpy as np
from jina.executors.indexers import BaseIndexer

from .. import FaissIndexer

# fix the seed here
np.random.seed(500)
retr_idx = None
vec_idx = np.random.randint(0, high=100, size=[10])
vec = np.array(np.random.random([10, 10]), dtype=np.float32)
query = np.array(np.random.random([10, 10]), dtype=np.float32)
cur_dir = os.path.dirname(os.path.abspath(__file__))


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)


def test_faiss_indexer():
    train_filepath = os.path.join(cur_dir, 'train.tgz')
    train_data = np.array(np.random.random([1024, 10]), dtype=np.float32)
    with gzip.open(train_filepath, 'wb', compresslevel=1) as f:
        f.write(train_data.tobytes())

    with FaissIndexer(index_filename='faiss.test.gz', index_key='IVF10,PQ2', train_filepath=train_filepath) as indexer:
        indexer.add(vec_idx, vec)
        indexer.save()
        assert os.path.exists(indexer.index_abspath)
        index_abspath = indexer.index_abspath
        save_abspath = indexer.save_abspath

    with BaseIndexer.load(save_abspath) as indexer:
        assert isinstance(indexer, FaissIndexer)
        idx, dist = indexer.query(query, top_k=4)
        global retr_idx
        if retr_idx is None:
            retr_idx = idx
        else:
            np.testing.assert_almost_equal(retr_idx, idx)
        assert idx.shape == dist.shape
        assert idx.shape == (10, 4)

    rm_files([index_abspath, save_abspath, train_filepath])



import os
import pickle
import shutil

import numpy as np

from .. import FeatureAgglomerationEncoder
from jina.executors import BaseExecutor
from jina.executors.encoders.numeric import TransformEncoder

input_dim = 28
target_output_dim = 2


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)


def encoding_results(encoder):
    test_data = np.random.rand(10, input_dim)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (test_data.shape[0], target_output_dim)
    assert type(encoded_data) == np.ndarray


def save_and_load(encoder, requires_train_after_load):
    test_data = np.random.rand(10, input_dim)

    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)

    if not requires_train_after_load:
        encoder_loaded = BaseExecutor.load(encoder.save_abspath)
        # some models are not deterministic when training, so even with same training data, we cannot ensure
        # same encoding results
        encoded_data_test = encoder_loaded.encode(test_data)
        np.testing.assert_array_equal(
            encoded_data_test, encoded_data_control)


def save_and_load_config(encoder, requires_train_after_load, train_data):
    assert encoder is not None

    test_data = np.random.rand(10, input_dim)

    encoder.save_config()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)

    if requires_train_after_load:
        encoder_loaded.train(train_data)
        encoded_data_test = encoder_loaded.encode(test_data)
        assert encoded_data_test.shape == (10, target_output_dim)


def test_random_feature_agglomeration_encoder_train():
    train_data = np.random.rand(2000, input_dim)
    encoder = FeatureAgglomerationEncoder(output_dim=target_output_dim)
    encoder.train(train_data)
    encoding_results(encoder)
    save_and_load(encoder, True)
    save_and_load_config(encoder, True, train_data)
    rm_files([encoder.save_abspath, encoder.config_abspath])


def test_random_feature_agglomeration_encoder_load():
    train_data = np.random.rand(2000, input_dim)

    from sklearn.cluster import FeatureAgglomeration
    model = FeatureAgglomeration(n_clusters=target_output_dim)
    filename = 'feature_agglomeration_model.model'
    pickle.dump(model.fit(train_data), open(filename, 'wb'))

    encoder = TransformEncoder(model_path=filename)

    test_data = np.random.rand(10, input_dim)
    encoded_data = encoder.encode(test_data)
    transformed_data = model.transform(test_data)
    assert encoded_data.shape == (test_data.shape[0], target_output_dim)
    assert type(encoded_data) == np.ndarray
    np.testing.assert_almost_equal(transformed_data, encoded_data)

    save_and_load(encoder, False)
    save_and_load_config(encoder, False, train_data)

    rm_files([encoder.save_abspath, encoder.config_abspath, filename])


__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Tuple

import numpy as np
from jina.executors.indexers.vector import BaseNumpyIndexer


class NmsLibIndexer(BaseNumpyIndexer):
    """nmslib powered vector indexer

    For documentation and explanation of each parameter, please refer to

        - https://nmslib.github.io/nmslib/quickstart.html
        - https://github.com/nmslib/nmslib/blob/master/manual/methods.md

    .. note::
        Nmslib package dependency is only required at the query time.
    """

    def __init__(self, space: str = 'cosinesimil', method: str = 'hnsw', print_progress: bool = False,
                 num_threads: int = 1,
                 *args, **kwargs):
        """
        Initialize an NmslibIndexer

        :param space: The metric space to create for this index
        :param method: The index method to use
        :param num_threads: The number of threads to use
        :param print_progress: Whether or not to display progress bar when creating index
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.method = method
        self.space = space
        self.print_progress = print_progress
        self.num_threads = num_threads

    def build_advanced_index(self, vecs: 'np.ndarray'):
        import nmslib
        _index = nmslib.init(method=self.method, space=self.space)
        _index.addDataPointBatch(vecs.astype(np.float32))
        _index.createIndex({'post': 2}, print_progress=self.print_progress)
        return _index

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        # if keys.dtype != np.float32:
        #     raise ValueError('vectors should be ndarray of float32')
        ret = self.query_handler.knnQueryBatch(keys, k=top_k, num_threads=self.num_threads)
        idx, dist = zip(*ret)
        return self.int2ext_key[np.array(idx)], np.array(dist)



import numpy as np

from .. import ImageFlipper


def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


def test_horizontal_flip():
    img_size = 217
    crafter = ImageFlipper()
    # generates a random image array of size (217, 217)
    img_array = create_random_img_array(img_size, img_size)
    crafted_chunk = crafter.craft(img_array)
    # image flips along the second axis (horizontal flip)
    flip_img_array = np.fliplr(img_array)
    # assert flipped image using numpy's fliplr method
    np.testing.assert_equal(crafted_chunk['blob'], flip_img_array)


def test_vertical_flip():
    img_size = 217
    crafter = ImageFlipper(vertical=True)
    # generates a random image array of size (217, 217)
    img_array = create_random_img_array(img_size, img_size)
    crafted_chunk = crafter.craft(img_array)
    # image flips along the first axis (vertical flip)
    flip_img_array = np.flipud(img_array)
    # assert flipped image using numpy's flipud method
    np.testing.assert_equal(crafted_chunk['blob'], flip_img_array)



from .. import ImageResizer


def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


def test_resize():
    img_width = 20
    img_height = 17
    output_dim = 71
    crafter = ImageResizer(target_size=output_dim)
    img_array = create_random_img_array(img_width, img_height)
    crafted_doc = crafter.craft(img_array)
    assert min(crafted_doc['blob'].shape[:-1]) == output_dim



import numpy as np

from .. import ArrayBytesReader


def test_bytes_reader():
    size = 8
    sample_array = np.random.rand(size).astype('float32')
    array_bytes = sample_array.tobytes()

    reader = ArrayBytesReader()
    crafted_doc = reader.craft(array_bytes, 0)

    assert crafted_doc['blob'].shape[0] == size
    np.testing.assert_array_equal(crafted_doc['blob'], sample_array)

def test_bytes_reader_int_type():
    size = 8
    sample_array = np.random.rand(size).astype('int')
    array_bytes = sample_array.tobytes()

    reader = ArrayBytesReader(as_type='int')
    crafted_doc = reader.craft(array_bytes, 0)

    assert crafted_doc['blob'].shape[0] == size
    np.testing.assert_array_equal(crafted_doc['blob'], sample_array)

def test_bytes_reader_wrong_type():
    size = 8
    sample_array = np.random.rand(size).astype('float32')
    array_bytes = sample_array.tobytes()

    reader = ArrayBytesReader(as_type='float64')
    crafted_doc = reader.craft(array_bytes, 0)

    assert crafted_doc['blob'].shape[0] == size / 2



import numpy as np
import os
import pickle
import shutil

from jina.executors.encoders.numeric import TransformEncoder
from jina.executors import BaseExecutor
from .. import FastICAEncoder

input_dim = 28
target_output_dim = 2
train_data = np.random.rand(2000, input_dim)

def rm_files(tmp_files):
    for file in tmp_files:
        if file and os.path.exists(file):
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.isdir(file):
                shutil.rmtree(file, ignore_errors=False, onerror=None)


def test_FastICATestCaseTrainCase():
    requires_train_after_load = True
    encoder = FastICAEncoder(
        output_dim=target_output_dim, whiten=True, num_features=input_dim, max_iter=200)
    encoder.train(train_data)
    encoding_results(encoder)
    save_and_load(encoder, requires_train_after_load)
    save_and_load_config(encoder, requires_train_after_load)
    rm_files([encoder.save_abspath, encoder.config_abspath, encoder.model_path])

def test_FastICATestCaseLoadCase():
    requires_train_after_load = False
    encoder = FastICAEncoder(
        output_dim=target_output_dim, whiten=True, num_features=input_dim, max_iter=200)
    encoder.train(train_data)
    filename = 'ica_model.model'
    pickle.dump(encoder.model, open(filename, 'wb'))
    encoder = TransformEncoder(model_path=filename)
    encoding_results(encoder)
    save_and_load(encoder, requires_train_after_load)
    save_and_load_config(encoder, requires_train_after_load)
    rm_files([encoder.save_abspath, encoder.config_abspath, encoder.model_path])


def encoding_results(encoder):
    test_data = np.random.rand(10, input_dim)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (test_data.shape[0], target_output_dim)
    assert type(encoded_data) is np.ndarray



def save_and_load(encoder, requires_train_after_load):
    test_data = np.random.rand(10, input_dim)
    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)

    if not requires_train_after_load:
        # some models are not deterministic when training, so even with same training data, we cannot ensure
        # same encoding results
        encoded_data_test = encoder_loaded.encode(test_data)
        np.testing.assert_array_equal(
            encoded_data_test, encoded_data_control)



def save_and_load_config(encoder, requires_train_after_load):
    test_data = np.random.rand(10, input_dim)
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)

    if requires_train_after_load:
        encoder_loaded.train(train_data)


    encoded_data_test = encoder_loaded.encode(test_data)
    assert encoded_data_test.shape == (10, target_output_dim)




import os
import mock
import shutil
import numpy as np
import pytest

from .. import VideoPaddleEncoder
from jina.executors import BaseExecutor
from jina.executors.metas import get_default_metas

input_dim = 224
target_output_dim = 2048
num_doc = 2
test_data = np.random.rand(num_doc, 3, 3, input_dim, input_dim)
tmp_files = []

def teardown():
    for k in tmp_files:
        if os.path.exists(k):
            if os.path.isfile(k):
                os.remove(k)
            elif os.path.isdir(k):
                shutil.rmtree(k, ignore_errors=False, onerror=None)

def add_tmpfile(*path):
    tmp_files.extend(path)


def get_encoder():
    metas = get_default_metas()
    if 'JINA_TEST_GPU' in os.environ:
        metas['on_gpu'] = True
    return VideoPaddleEncoder(metas=metas)


class MockModule:
    def context(self, *args, **kwargs):
        input_name = MockModel()
        input_name.name="video"
        return [input_name], {'feature_map': MockTensor('output')}, MockModel()


class MockTensor:
    def __init__(self, name):
        self.name = name


class MockModel:
    pass


class MockExecutor:
    def __init__(self, *args, **kwargs):
        pass

    def close(self):
        pass

    def run(self, feed, *args, **kwargs):
        feature_map = np.random.random((num_doc, target_output_dim))
        return feature_map, None


class MockCUDADevice:
    pass


class MockCPUDevice:
    pass


def _test_videopaddlehubencoder_encode(*args, **kwargs):
    encoder = get_encoder()
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (num_doc, target_output_dim)
    add_tmpfile(encoder.save_abspath, encoder.config_abspath)
    teardown()


@mock.patch('paddlehub.Module', return_value=MockModule())
@mock.patch('paddle.fluid.Executor', return_value=MockExecutor())
@mock.patch('paddle.fluid.CUDAPlace', return_value=MockCUDADevice())
@mock.patch('paddle.fluid.CPUPlace', return_value=MockCPUDevice())
def test_videopaddlehubencoder_encode(*args, **kwargs):
    _test_videopaddlehubencoder_encode(*args, **kwargs)


@mock.patch('paddlehub.Module', return_value=MockModule())
@mock.patch('paddle.fluid.Executor', return_value=MockExecutor())
@mock.patch('paddle.fluid.CUDAPlace', return_value=MockCUDADevice())
@mock.patch('paddle.fluid.CPUPlace', return_value=MockCPUDevice())
def test_videopaddlehubencoder_save_and_load(*args, **kwargs):
    encoder = get_encoder()
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)
    assert encoder_loaded.model_name == encoder.model_name
    add_tmpfile(encoder.save_abspath, encoder.config_abspath)
    teardown()


@mock.patch('paddlehub.Module', return_value=MockModule())
@mock.patch('paddle.fluid.Executor', return_value=MockExecutor())
@mock.patch('paddle.fluid.CUDAPlace', return_value=MockCUDADevice())
@mock.patch('paddle.fluid.CPUPlace', return_value=MockCPUDevice())
def test_videopaddlehubencoder_save_and_load_config(*args, **kwargs):
    encoder = get_encoder()
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    assert encoder_loaded.model_name == encoder.model_name
    add_tmpfile(encoder.save_abspath, encoder.config_abspath)
    teardown()


@pytest.mark.skipif('JINA_TEST_PRETRAINED' not in os.environ, reason='skip the pretrained test if not set')
def test_videopaddlehubencoder_encode_with_pretrained_model():
    _test_videopaddlehubencoder_encode()


from .. import RandomImageCropper


def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


def test_random_crop():
    img_size = 217
    img_array = create_random_img_array(img_size, img_size)
    output_dim = 20
    num_patches = 20
    crafter = RandomImageCropper(output_dim, num_patches)
    chunks = crafter.craft(img_array)
    assert len(chunks) == num_patches
    for chunk in chunks:
        assert chunk['location'][0] <= (img_size - output_dim)
        assert chunk['location'][1] <= (img_size - output_dim)



import os
import shutil
import numpy as np
from .. import BigTransferEncoder
from jina.executors.metas import get_default_metas
from jina.executors import BaseExecutor


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)


def get_encoder():
    metas = get_default_metas()
    if 'JINA_TEST_GPU' in os.environ:
        metas['on_gpu'] = True
    return BigTransferEncoder(model_path='pretrained', channel_axis=1, metas=metas)




# def test_encoding_results():
#     input_dim = 48
#     output_dim = 2048
#     encoder = get_encoder()
#     test_data = np.random.rand(2, 3, input_dim, input_dim)
#     encoded_data = encoder.encode(test_data)
#     assert encoded_data.shape == (2, output_dim)
#     rm_files([encoder.save_abspath, encoder.config_abspath])
#
#
# def test_save_and_load():
#     encoder = get_encoder()
#     encoder.touch()
#     encoder.save()
#     assert os.path.exists(encoder.save_abspath)
#     encoder_loaded = BaseExecutor.load(encoder.save_abspath)
#     assert encoder_loaded.channel_axis == encoder.channel_axis
#     rm_files([encoder.save_abspath, encoder.config_abspath])


def test_save_and_load_config():
    encoder = get_encoder()
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    assert encoder_loaded.model_path == encoder.model_path
    # rm_files([encoder.save_abspath, encoder.config_abspath])


__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import mimetypes
import os
import urllib.parse
from typing import Iterator, Union

import numpy as np

from ...counter import RandomUintCounter, SimpleCounter
from ...drivers.helper import array2pb, guess_mime
from ...enums import ClientMode
from ...helper import batch_iterator, is_url
from ...logging import default_logger
from ...proto import jina_pb2

if False:
    from ...counter import BaseCounter


def _add_document(request: 'jina_pb2.Request', content: Union['jina_pb2.Document', 'np.ndarray', bytes, str], mode: str,
                  doc_counter: 'BaseCounter', docs_in_same_batch: int, mime_type: str, buffer_sniff: bool,
                  granularity: int):
    d = getattr(request, str(mode).lower()).docs.add()
    if isinstance(content, jina_pb2.Document):
        d.CopyFrom(content)
    elif isinstance(content, np.ndarray):
        d.blob.CopyFrom(array2pb(content))
    elif isinstance(content, bytes):
        d.buffer = content
        if not mime_type and buffer_sniff:
            try:
                import magic
                mime_type = magic.from_buffer(content, mime=True)
            except Exception as ex:
                default_logger.warning(f'can not sniff the MIME type due to the exception {ex}')
    elif isinstance(content, str):
        scheme = urllib.parse.urlparse(content).scheme
        if ((scheme in {'http', 'https'} and is_url(content)) or (scheme in {'data'}) or os.path.exists(content)
                or os.access(os.path.dirname(content), os.W_OK)):
            d.uri = content
            mime_type = guess_mime(content)
        else:
            d.text = content
            mime_type = 'text/plain'
    else:
        raise TypeError(f'{type(content)} type of input is not supported')

    if mime_type:
        d.mime_type = mime_type

    d.id = next(doc_counter)
    d.weight = 1.0
    d.length = docs_in_same_batch
    d.granularity = granularity


def _generate(data: Union[Iterator['jina_pb2.Document'], Iterator[bytes], Iterator['np.ndarray'], Iterator[str], 'np.ndarray'],
              batch_size: int = 0, first_doc_id: int = 0, first_request_id: int = 0,
              random_doc_id: bool = False, mode: ClientMode = ClientMode.INDEX,
              mime_type: str = None, queryset: Iterator['jina_pb2.QueryLang'] = None,
              granularity: int = 0, *args, **kwargs) -> Iterator['jina_pb2.Message']:
    buffer_sniff = False
    doc_counter = RandomUintCounter() if random_doc_id else SimpleCounter(first_doc_id)
    req_counter = SimpleCounter(first_request_id)

    try:
        import magic
        buffer_sniff = True
    except (ImportError, ModuleNotFoundError):
        default_logger.warning(f'can not sniff the MIME type '
                               f'MIME sniffing requires pip install "jina[http]" '
                               f'and brew install libmagic (Mac)/ apt-get install libmagic1 (Linux)')

    if mime_type and (mime_type not in mimetypes.types_map.values()):
        mime_type = mimetypes.guess_type(f'*.{mime_type}')[0]

    if isinstance(mode, str):
        mode = ClientMode.from_string(mode)

    for batch in batch_iterator(data, batch_size):
        req = jina_pb2.Request()
        req.request_id = next(req_counter)
        if queryset:
            if isinstance(queryset, jina_pb2.QueryLang):
                queryset = [queryset]
            req.queryset.extend(queryset)

        for content in batch:
            _add_document(request=req, content=content, mode=mode, doc_counter=doc_counter,
                          docs_in_same_batch=batch_size, mime_type=mime_type,
                          buffer_sniff=buffer_sniff, granularity=granularity)
        yield req


def index(*args, **kwargs):
    """Generate a indexing request"""
    yield from _generate(*args, **kwargs)


def train(*args, **kwargs):
    """Generate a training request """
    yield from _generate(*args, **kwargs)
    req = jina_pb2.Request()
    req.request_id = 1
    req.train.flush = True
    yield req


def search(*args, **kwargs):
    """Generate a searching request """
    yield from _generate(*args, **kwargs)



import tempfile
import os

import numpy as np

from .. import CustomKerasImageEncoder
from jina.executors.metas import get_default_metas
from jina.executors import BaseExecutor
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, Activation, Flatten, Dense


def rm_files(file_paths):
    for file_path in file_paths:
        if os.path.exists(file_path):
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path, ignore_errors=False, onerror=None)

class TestNet:
    def __init__(self):
        self.model = None
        self.input_shape = (224, 224, 3)
        self.conv = Conv2D(32, (3, 3), padding='same', name='conv1', input_shape=self.input_shape)
        self.activation_relu = Activation('relu')
        self.flatten = Flatten()
        self.dense = Dense(10, name='dense')
        self.activation_softmax = Activation('softmax')

    def create_model(self):
        self.model = Sequential()
        self.model.add(self.conv)
        self.model.add(self.activation_relu)
        self.model.add(self.flatten)
        self.model.add(self.dense)
        self.model.add(self.activation_softmax)
        return self.model

def get_encoder():
    metas = get_default_metas()
    if 'JINA_TEST_GPU' in os.environ:
        metas['on_gpu'] = True
    path = tempfile.NamedTemporaryFile().name
    model = TestNet().create_model()
    model.save(path)
    return CustomKerasImageEncoder(channel_axis=1, model_path=path, layer_name='dense', metas=metas)

def test_encoding_results():
    target_output_dim = 10
    input_dim = 224
    encoder = get_encoder()
    test_data = np.random.rand(2, 3, input_dim, input_dim)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (2, target_output_dim)

def test_save_and_load():
    input_dim = 224
    encoder = get_encoder()
    test_data = np.random.rand(2, 3, input_dim, input_dim)
    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)
    encoded_data_test = encoder_loaded.encode(test_data)
    assert encoder_loaded.channel_axis == encoder.channel_axis
    np.testing.assert_array_equal(encoded_data_control, encoded_data_test)

def test_save_and_load_config():
    encoder = get_encoder()
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    assert encoder_loaded.channel_axis == encoder.channel_axis


from .. import MFCCTimbreEncoder
import numpy as np



def test_mfcc_encoder():
    batch_size = 10
    n_frames = 5
    signal_length = 500 * n_frames
    test_data = np.random.randn(batch_size, signal_length)
    n_mfcc = 12
    encoder = MFCCTimbreEncoder(n_mfcc=n_mfcc)
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (batch_size, n_mfcc * n_frames)



from .. import SlidingWindowImageCropper


def create_random_img_array(img_height, img_width):
    import numpy as np
    return np.random.randint(0, 256, (img_height, img_width, 3))


def test_sliding_windows_no_padding():
    img_size = 14
    img_array = create_random_img_array(img_size, img_size)
    output_dim = 4
    strides = (6, 6)
    crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=False)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 4
    assert chunks[0]['location'] == (0, 0)
    assert chunks[1]['location'] == (0, 6)
    assert chunks[2]['location'] == (6, 0)
    assert chunks[3]['location'] == (6, 6)


def test_sliding_windows_with_padding():
    img_size = 14
    img_array = create_random_img_array(img_size, img_size)
    output_dim = 4
    strides = (6, 6)
    crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=True)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 9
    assert chunks[0]['location'] == (0, 0)
    assert chunks[1]['location'] == (0, 6)
    assert chunks[2]['location'] == (0, 12)
    assert chunks[3]['location'] == (6, 0)
    assert chunks[4]['location'] == (6, 6)
    assert chunks[5]['location'] == (6, 12)
    assert chunks[6]['location'] == (12, 0)
    assert chunks[7]['location'] == (12, 6)
    assert chunks[8]['location'] == (12, 12)


def test_sliding_windows_without_padding_rectangular_ugly_shapes():
    height = 16
    width = 11
    img_array = create_random_img_array(img_height=height, img_width=width)
    output_dim = 4
    strides = (4, 4)
    crafter = SlidingWindowImageCropper(target_size=output_dim, strides=strides, padding=False)
    chunks = crafter.craft(img_array)
    assert len(chunks) == 8
    assert chunks[0]['location'] == (0, 0)
    assert chunks[1]['location'] == (0, 4)
    assert chunks[2]['location'] == (4, 0)
    assert chunks[3]['location'] == (4, 4)
    assert chunks[4]['location'] == (8, 0)
    assert chunks[5]['location'] == (8, 4)
    assert chunks[6]['location'] == (12, 0)
    assert chunks[7]['location'] == (12, 4)



import os
import shutil
import numpy as np
import pytest
from .. import TransformerTorchEncoder
from jina.executors import BaseExecutor
from jina.executors.metas import get_default_metas


def rm_files(tmp_files):
    for k in tmp_files:
        if os.path.exists(k):
            if os.path.isfile(k):
                os.remove(k)
            elif os.path.isdir(k):
                shutil.rmtree(k, ignore_errors=False, onerror=None)


os.environ['TEST_WORKDIR'] = os.getcwd()

metas = get_default_metas()
if 'JINA_TEST_GPU' in os.environ:
    metas['on_gpu'] = True

encoders = [
    TransformerTorchEncoder(
        pretrained_model_name_or_path='bert-base-uncased',
        model_save_path='bert-base-uncased',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='mean',
        pretrained_model_name_or_path='bert-base-uncased',
        model_save_path='bert-base-uncased-mean',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='min',
        pretrained_model_name_or_path='bert-base-uncased',
        model_save_path='bert-base-uncased-min',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='max',
        pretrained_model_name_or_path='bert-base-uncased',
        model_save_path='bert-base-uncased-max',
        metas=metas),
    TransformerTorchEncoder(
        pretrained_model_name_or_path='xlnet-base-cased',
        model_save_path='xlnet-base-cased',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='mean',
        pretrained_model_name_or_path='xlnet-base-cased',
        model_save_path='xlnet-base-cased-mean',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='min',
        pretrained_model_name_or_path='xlnet-base-cased',
        model_save_path='xlnet-base-cased-min',
        metas=metas),
    TransformerTorchEncoder(
        pooling_strategy='max',
        pretrained_model_name_or_path='xlnet-base-cased',
        model_save_path='xlnet-base-cased-max',
        metas=metas),
]

@pytest.mark.parametrize('encoder', encoders)
def test_encoding_results(encoder):
    target_output_dim = 768
    test_data = np.array(['it is a good day!', 'the dog sits on the floor.'])
    encoded_data = encoder.encode(test_data)
    assert encoded_data.shape == (2, target_output_dim)
    assert not np.allclose(encoded_data[0], encoded_data[1])

@pytest.mark.parametrize('encoder', encoders)
def test_save_and_load(encoder):
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    test_data = np.array(['a', 'b', 'c', 'x', '!'])
    encoded_data_control = encoder.encode(test_data)
    encoder.touch()
    encoder.save()
    assert os.path.exists(encoder.save_abspath)
    encoder_loaded = BaseExecutor.load(encoder.save_abspath)
    encoded_data_test = encoder_loaded.encode(test_data)
    assert encoder_loaded.max_length == encoder.max_length
    np.testing.assert_array_equal(encoded_data_control, encoded_data_test)
    rm_files([encoder.config_abspath, encoder.save_abspath, encoder_loaded.config_abspath, encoder_loaded.save_abspath])


@pytest.mark.parametrize('encoder', encoders)
def test_save_and_load_config(encoder):
    encoder.save_config()
    assert os.path.exists(encoder.config_abspath)
    encoder_loaded = BaseExecutor.load_config(encoder.config_abspath)
    assert encoder_loaded.max_length == encoder.max_length
    rm_files([encoder_loaded.config_abspath, encoder_loaded.save_abspath])
